#pragma once

#define LOCK_VERSION 3

#if LOCK_VERSION == 1
class CObject;

class CProxy
{
	friend class CObject;
public:
	__declspec(property(get=get_Value,put=put_Value)) int value;
	int get_Value(int level = -1) const;
	void put_Value(int i);

	void Commit();
	void Rollback();

private:	
	int _level;
	int _value;

	bool fUpd;

	CProxy(CObject* par,int level)
	{
		fUpd = false;
		parent = par;
		_level = level;
	}

	CObject* parent;
};

class CObject
{
	friend class CProxy;
public:

	static CProxy& GetObject(int level = -1);

	~CObject()
	{
		if (hMutex) CloseHandle(hMutex);
	}

protected:
	CProxy& BeginTran(int level)
	{
		return *(new CProxy(this,level));
	}

	void RequestExclusive(int level)
	{
	}
	
	void RequestShared(int level)
	{
	}

	void RemoveShared(int level)
	{
	}

	void RemoveLocks()
	{
	}

private:
	CObject()
	{
		value = 0;
	}

	int value;
	static HANDLE hMutex;
};

__declspec(selectany) HANDLE CObject::hMutex = NULL;

CProxy& CObject::GetObject(int level)
{
	HANDLE hLocMutex = CreateMutex(NULL,TRUE,_T("Guard-Lock-Mutex"));
	bool flg = GetLastError() == ERROR_ALREADY_EXISTS;
	
	if (flg) WaitForSingleObject(hLocMutex,INFINITE);
	else CObject::hMutex = hLocMutex;

	static CObject obj;

	ReleaseMutex(hLocMutex);
	
	if (flg) CloseHandle(hLocMutex);

	return obj.BeginTran(level);

}

void CProxy::Commit()
{
	parent->RemoveLocks();
	delete this;
}

void CProxy::Rollback()
{
	if (fUpd)
		parent->value = _value;
	parent->RemoveLocks();
	delete this;
}

void CProxy::put_Value(int i)
{
	parent->RequestExclusive(_level);
	if (!fUpd)
		_value = parent->value;
	parent->value = i;
	fUpd = true;
}

int CProxy::get_Value(int level) const
{
	if (level == -1) 
		level = _level;

	parent->RequestShared(level);
	int v = parent->value;
	parent->RemoveShared(level);
	return v;
}

//LOCK_VERSION == 1
#elif LOCK_VERSION == 2

class CObject;

class CProxy
{
	friend class CObject;
public:
	__declspec(property(get=get_Value,put=put_Value)) int value;
	int get_Value(int level = -1) const;
	void put_Value(int i);

	void Commit();
	void Rollback();

private:	
	int _level;
	int _value;

	bool fUpd;

	CProxy(CObject* par,int level)
	{
		fUpd = false;
		parent = par;
		_level = level;
	}

	CObject* parent;
};

class CObject
{
	friend class CProxy;
public:

	enum {READ_UNCOMMITTED};

	static CProxy& GetObject(int level = -1);

	~CObject()
	{
		DeleteCriticalSection(&exclusive);
		if (hMutex) CloseHandle(hMutex);
	}

protected:
	CProxy& BeginTran(int level)
	{
		return *(new CProxy(this,level));
	}

	void RequestExclusive(int level)
	{
		if (level >= READ_UNCOMMITTED)
			TestExclusive();
	}
	
	void RequestShared(int level)
	{
	}

	void RemoveShared(int level)
	{
	}

	void RemoveLocks()
	{
		RemoveAllLocks();
	}

private:
	CObject()
	{
		value = 0;
		InitializeCriticalSection(&exclusive);
	}

	void TestExclusive()
	{
		//   
		EnterCriticalSection(&exclusive);

		//   
		if (exclusive.RecursionCount > 1)
			LeaveCriticalSection(&exclusive);
	}

	void RemoveAllLocks()
	{
		//     - 
		if (exclusive.OwningThread == (HANDLE)GetCurrentThreadId())
			LeaveCriticalSection(&exclusive);
	}

	int value;
	CRITICAL_SECTION exclusive;
	static HANDLE hMutex;
};

__declspec(selectany) HANDLE CObject::hMutex = NULL;

CProxy& CObject::GetObject(int level)
{
	HANDLE hLocMutex = CreateMutex(NULL,TRUE,_T("Guard-Lock-Mutex"));
	bool flg = GetLastError() == ERROR_ALREADY_EXISTS;
	
	if (flg) WaitForSingleObject(hLocMutex,INFINITE);
	else CObject::hMutex = hLocMutex;

	static CObject obj;

	ReleaseMutex(hLocMutex);
	
	if (flg) CloseHandle(hLocMutex);

	return obj.BeginTran(level);

}

void CProxy::Commit()
{
	parent->RemoveLocks();
	delete this;
}

void CProxy::Rollback()
{
	if (fUpd)
		parent->value = _value;
	parent->RemoveLocks();
	delete this;
}

void CProxy::put_Value(int i)
{
	parent->RequestExclusive(_level);
	if (!fUpd)
		_value = parent->value;
	parent->value = i;
	fUpd = true;
}

int CProxy::get_Value(int level) const
{
	if (level == -1) 
		level = _level;

	parent->RequestShared(level);
	int v = parent->value;
	parent->RemoveShared(level);
	return v;
}

//LOCK_VERSION == 2
#elif LOCK_VERSION == 3

class CObject;

class CProxy
{
	friend class CObject;
public:
	__declspec(property(get=get_Value,put=put_Value)) int value;
	int get_Value(int level = -1) const;
	void put_Value(int i);

	void Commit();
	void Rollback();

private:	
	int _level;
	int _value;

	bool fUpd;

	CProxy(CObject* par,int level)
	{
		fUpd = false;
		parent = par;
		_level = level;
	}

	CObject* parent;
};

class CObject
{
	friend class CProxy;
public:

	enum {READ_UNCOMMITTED,READ_COMMITTED,REPEATABLE_READ};

	static CProxy& GetObject(int level = -1);

	~CObject()
	{
		DeleteCriticalSection(&exclusive);
		if (hShared) CloseHandle(hShared);
		if (hMutex) CloseHandle(hMutex);
	}

protected:
	CProxy& BeginTran(int level)
	{
		return *(new CProxy(this,level));
	}

	void RequestExclusive(int level)
	{
		if (level >= READ_UNCOMMITTED)
			TestExclusive();
	}
	
	void RequestShared(int level)
	{
		if (level > READ_UNCOMMITTED)
			TestShared(level);
	}

	void RemoveShared(int level)
	{
		if (level == READ_COMMITTED){
			RemoveSharedLock();			
		}
	}

	void RemoveLocks()
	{
		RemoveAllLocks();
	}

private:
	CObject()
	{
		value = 0;
		InitializeCriticalSection(&exclusive);
		hShared = CreateEvent(NULL,FALSE,TRUE,NULL);
	}

	void TestShared(int level)
	{
		//   
		EnterCriticalSection(&exclusive);

		//  
		//      
		if (exclusive.RecursionCount == 1)
			ResetEvent(hShared);

		//  
		LeaveCriticalSection(&exclusive);
	}

	void TestExclusive()
	{
		//   
		WaitForSingleObject(hShared,INFINITE);

		//   
		EnterCriticalSection(&exclusive);

		//   
		if (exclusive.RecursionCount > 1)
			LeaveCriticalSection(&exclusive);
	}

	void RemoveSharedLock()
	{
		SetEvent(hShared);
	}

	void RemoveAllLocks()
	{
		RemoveSharedLock();

		//     - 
		if (exclusive.OwningThread == (HANDLE)GetCurrentThreadId())
			LeaveCriticalSection(&exclusive);
	}

	int value;
	CRITICAL_SECTION exclusive;
	HANDLE hShared;
	static HANDLE hMutex;
};

__declspec(selectany) HANDLE CObject::hMutex = NULL;

CProxy& CObject::GetObject(int level)
{
	HANDLE hLocMutex = CreateMutex(NULL,TRUE,_T("Guard-Lock-Mutex"));
	bool flg = GetLastError() == ERROR_ALREADY_EXISTS;
	
	if (flg) WaitForSingleObject(hLocMutex,INFINITE);
	else CObject::hMutex = hLocMutex;

	static CObject obj;

	ReleaseMutex(hLocMutex);
	
	if (flg) CloseHandle(hLocMutex);

	return obj.BeginTran(level);

}

void CProxy::Commit()
{
	parent->RemoveLocks();
	delete this;
}

void CProxy::Rollback()
{
	if (fUpd)
		parent->value = _value;
	parent->RemoveLocks();
	delete this;
}

void CProxy::put_Value(int i)
{
	parent->RequestExclusive(_level);
	if (!fUpd)
		_value = parent->value;
	parent->value = i;
	fUpd = true;
}

int CProxy::get_Value(int level) const
{
	if (level == -1) 
		level = _level;

	parent->RequestShared(level);
	int v = parent->value;
	parent->RemoveShared(level);
	return v;
}

//LOCK_VERSION == 3
#elif LOCK_VERSION == 10

class CObject;

class CProxy
{
	friend class CObject;
public:
	__declspec(property(get=get_Value,put=put_Value)) int value;
	int get_Value(int level = -1) const;
	void put_Value(int i);

	void Commit();
	void Rollback();

private:	
	int _level;
	int _value;

	bool fUpd;

	CProxy(CObject* par,int level)
	{
		fUpd = false;
		parent = par;
		_level = level;
	}

	CObject* parent;
};

class CObject
{
	friend class CProxy;
public:

	enum {
#ifdef MSSQL
		READ_UNCOMMITTED,
#elif defined ORACLE
		SNAP_SHOT,
#endif
		READ_COMMITTED,REPEATABLE_READ,UPDLOCK,XLOCK};

	static CProxy& GetObject(int level = -1);

	~CObject()
	{
		DeleteCriticalSection(&exclusive);
		DeleteCriticalSection(&update);
		if (hMutex) CloseHandle(hMutex);
		if (hShared) CloseHandle(hShared);
	}

protected:
	CProxy& BeginTran(int level)
	{
		return *(new CProxy(this,level));
	}

	void RequestExclusive(int level)
	{
		ATLASSERT(level <= REPEATABLE_READ);
#ifdef MSSQL
		if (level >= READ_UNCOMMITTED)
#elif defined ORACLE
		if (level >= SNAP_SHOT)
#endif
			TestExclusive();
	}
	
	void RequestShared(int level)
	{

#ifdef MSSQL
		if (level > READ_UNCOMMITTED)
#elif defined ORACLE
		if (level > SNAP_SHOT)
#endif
			TestShared(level);
	}

	void RemoveShared(int level)
	{
		if (level == READ_COMMITTED){
			RemoveSharedLock();			
		}
	}

	void RemoveLocks()
	{
		RemoveAllLocks();
	}

private:
	CObject()
	{
		value = 0;
		InitializeCriticalSection(&update);
		InitializeCriticalSection(&exclusive);
		hShared = CreateEvent(NULL,FALSE,TRUE,NULL);
	}

	void TestShared(int level)
	{
		//   
		EnterCriticalSection(&exclusive);

		//  
		if (level == UPDLOCK){
			EnterCriticalSection(&update);

			//   
			if (update.RecursionCount > 1)
				LeaveCriticalSection(&update);
		}
		else if (level != XLOCK){
		
			//  
			//       
			// 
			if (update.OwningThread != (HANDLE)GetCurrentThreadId() &&
				exclusive.RecursionCount == 1)
				ResetEvent(hShared);

			//  
			LeaveCriticalSection(&exclusive);
		}
		//  XLOCK   

	}

	void TestExclusive()
	{
		//   
		WaitForSingleObject(hShared,INFINITE);

		//   
		EnterCriticalSection(&update);

		//   
		EnterCriticalSection(&exclusive);

		//  
		LeaveCriticalSection(&update);

		//   
		if (exclusive.RecursionCount > 1)
			LeaveCriticalSection(&exclusive);
	}

	void RemoveSharedLock()
	{
		SetEvent(hShared);
	}

	void RemoveAllLocks()
	{
		RemoveSharedLock();
		
		//     - 
		if (update.OwningThread == (HANDLE)GetCurrentThreadId())
			LeaveCriticalSection(&update);

		//     - 
		if (exclusive.OwningThread == (HANDLE)GetCurrentThreadId())
			LeaveCriticalSection(&exclusive);
	}

	int value;
	CRITICAL_SECTION update;
	CRITICAL_SECTION exclusive;
	HANDLE hShared;
	static HANDLE hMutex;
};

__declspec(selectany) HANDLE CObject::hMutex = NULL;

CProxy& CObject::GetObject(int level)
{
	HANDLE hLocMutex = CreateMutex(NULL,TRUE,_T("Guard-Lock-Mutex"));
	bool flg = GetLastError() == ERROR_ALREADY_EXISTS;
	
	if (flg) WaitForSingleObject(hLocMutex,INFINITE);
	else CObject::hMutex = hLocMutex;

	static CObject obj;

	ReleaseMutex(hLocMutex);
	
	if (flg) CloseHandle(hLocMutex);

	return obj.BeginTran(level);

}

void CProxy::Commit()
{
#ifdef ORACLE
	parent->value = _value;
#endif
	parent->RemoveLocks();
	delete this;
}

void CProxy::Rollback()
{
#ifdef MSSQL
	if (fUpd)
		parent->value = _value;
#endif
	parent->RemoveLocks();
	delete this;
}

void CProxy::put_Value(int i)
{
	parent->RequestExclusive(_level);
#ifdef MSSQL
	if (!fUpd)
		_value = parent->value;
	parent->value = i;
#elif defined ORACLE
	_value = i;
#endif
	fUpd = true;
}

int CProxy::get_Value(int level) const
{
	if (level == -1) 
		level = _level;

	parent->RequestShared(level);
#ifdef MSSQL
	int v = parent->value;
	parent->RemoveShared(level);
	return v;
#elif defined ORACLE
	return _value;
#endif
}

//LOCK_VERSION == 10
#endif